import glob
import json
import os
import shutil
import sys
import time
import unittest
import uuid

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
__package__ = "tests"
# from .context import LocalSession
# from .context import LocalContext
from .context import (
    PBS,
    Job,
    JobStatus,
    LocalContext,
    Resources,
    Submission,
    Task,
    setUpModule,
)
from .sample_class import SampleClass


class TestResources(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None
        self.resources = SampleClass.get_sample_resources()
        self.resources_dict = SampleClass.get_sample_resources_dict()

    def test_eq(self):
        self.assertEqual(self.resources, SampleClass.get_sample_resources())

    def test_serialize(self):
        self.assertEqual(self.resources.serialize(), self.resources_dict)

    def test_deserialize(self):
        resources = Resources.deserialize(resources_dict=self.resources_dict)
        self.assertEqual(self.resources, resources)

    def test_serialize_deserialize(self):
        self.assertEqual(
            self.resources,
            Resources.deserialize(resources_dict=self.resources.serialize()),
        )

    def test_resources_json(self):
        with open("jsons/resources.json", "r") as f:
            resources_json_dict = json.load(f)
        self.assertTrue(resources_json_dict, self.resources_dict)
        self.assertTrue(resources_json_dict, self.resources.serialize())

    def test_load_from_json(self):
        resources = Resources.load_from_json("jsons/resources.json")
        self.assertTrue(resources, self.resources)



# from dpdispatcher.batch_object import BatchObject
# from dpdispatcher.batch import Batch
import os
import shutil
import sys

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
__package__ = "tests"
import json
import unittest

from .context import (
    Job,
    LocalContext,
    Machine,
    Resources,
    Shell,
    Submission,
    Task,
    get_file_md5,
)


@unittest.skipIf(sys.platform == "win32", "Shell is not supported on Windows")
class TestShellTrival(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None
        # self.local_context_dict = {
        #     'local_root': './test_shell_trival_dir',
        #     'remote_root': './tmp_shell_trival_dir'
        # }

    def test_shell_trival(self):
        with open("jsons/machine_local_shell.json", "r") as f:
            machine_dict = json.load(f)

        machine = Machine(**machine_dict["machine"])
        resources = Resources(**machine_dict["resources"])

        task1 = Task(
            command="cat example.txt",
            task_work_path="dir1/",
            forward_files=["example.txt"],
            backward_files=["out.txt"],
            outlog="out.txt",
        )
        task2 = Task(
            command="cat example.txt",
            task_work_path="dir2/",
            forward_files=["example.txt"],
            backward_files=["out.txt"],
            outlog="out.txt",
        )
        task3 = Task(
            command="cat example.txt",
            task_work_path="dir3/",
            forward_files=["example.txt"],
            backward_files=["out.txt"],
            outlog="out.txt",
        )
        task4 = Task(
            command="cat example.txt",
            task_work_path="dir4/",
            forward_files=["example.txt"],
            backward_files=["out.txt"],
            outlog="out.txt",
        )
        task_list = [task1, task2, task3, task4]

        submission = Submission(
            work_base="parent_dir/",
            machine=machine,
            resources=resources,
            forward_common_files=["graph.pb"],
            backward_common_files=[],
            task_list=task_list,
        )
        submission.run_submission(clean=False)

        for dir in ["dir1", "dir2", "dir3", "dir4"]:
            f1 = os.path.join(
                "test_shell_trival_dir/", "parent_dir/", dir, "example.txt"
            )
            f2 = os.path.join("test_shell_trival_dir/", "parent_dir/", dir, "out.txt")
            self.assertEqual(get_file_md5(f1), get_file_md5(f2))

    def test_shell_fail(self):
        with open("jsons/machine_local_shell.json", "r") as f:
            machine_dict = json.load(f)

        machine = Machine(**machine_dict["machine"])
        resources = Resources(**machine_dict["resources"])

        task = Task(
            command="cat mock_fail_task.txt && exit 1",
            task_work_path="./",
            forward_files=[],
            backward_files=["out.txt"],
            outlog="out.txt",
        )

        task_list = [
            task,
        ]

        submission = Submission(
            work_base="fail_dir/",
            machine=machine,
            resources=resources,
            forward_common_files=[],
            backward_common_files=[],
            task_list=task_list,
        )
        with self.assertRaises(RuntimeError):
            submission.run_submission()

    def test_shell_recover(self):
        with open("jsons/machine_lazylocal_shell.json", "r") as f:
            machine_dict = json.load(f)

        machine = Machine(**machine_dict["machine"])
        resources = Resources(**machine_dict["resources"])

        pass

    def test_dir_with_space(self):
        """Test directory with space"""
        with open("jsons/machine_local_shell.json", "r") as f:
            machine_dict = json.load(f)

        machine = Machine(**machine_dict["machine"])
        resources = Resources(**machine_dict["resources"])

        task1 = Task(
            command="cat example.txt",
            task_work_path="dir with space/",
            forward_files=["example.txt"],
            backward_files=["out.txt"],
            outlog="out.txt",
        )
        task_list = [task1]

        submission = Submission(
            work_base="parent_dir/",
            machine=machine,
            resources=resources,
            forward_common_files=[],
            backward_common_files=[],
            task_list=task_list,
        )
        submission.run_submission(clean=False)

        for dir in ["dir with space"]:
            f1 = os.path.join(
                "test_shell_trival_dir/", "parent_dir/", dir, "example.txt"
            )
            f2 = os.path.join("test_shell_trival_dir/", "parent_dir/", dir, "out.txt")
            self.assertEqual(get_file_md5(f1), get_file_md5(f2))

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree("tmp_shell_trival_dir/")



import json

from dargs import ArgumentEncoder

from dpdispatcher import Machine, Resources, Task

resources_dargs = Resources.arginfo()
with open("dpdispatcher-resources.json", "w") as f:
    json.dump(resources_dargs, f, cls=ArgumentEncoder)

machine_dargs = Machine.arginfo()
with open("dpdispatcher-machine.json", "w") as f:
    json.dump(machine_dargs, f, cls=ArgumentEncoder)

task_dargs = Task.arginfo()
with open("dpdispatcher-task.json", "w") as f:
    json.dump(task_dargs, f, cls=ArgumentEncoder)



# from dpdispatcher.batch_object import BatchObject
# from dpdispatcher.batch import Batch
import os
import shutil
import sys
import textwrap

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
__package__ = "tests"
import json
import unittest

from .context import (
    Job,
    LocalContext,
    Machine,
    Resources,
    Shell,
    Submission,
    Task,
    get_file_md5,
)


class TestSlurmScriptGeneration(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None

    def test_shell_trival(self):
        with open("jsons/machine_lazy_local_slurm.json", "r") as f:
            machine_dict = json.load(f)

        machine = Machine(**machine_dict["machine"])
        resources = Resources(**machine_dict["resources"])

        task1 = Task(
            command="cat example.txt",
            task_work_path="dir1/",
            forward_files=["example.txt"],
            backward_files=["out.txt"],
            outlog="out.txt",
        )
        task2 = Task(
            command="cat example.txt",
            task_work_path="dir2/",
            forward_files=["example.txt"],
            backward_files=["out.txt"],
            outlog="out.txt",
        )
        task3 = Task(
            command="cat example.txt",
            task_work_path="dir3/",
            forward_files=["example.txt"],
            backward_files=["out.txt"],
            outlog="out.txt",
        )
        task4 = Task(
            command="cat example.txt",
            task_work_path="dir4/",
            forward_files=["example.txt"],
            backward_files=["out.txt"],
            outlog="out.txt",
        )
        task_list = [task1, task2, task3, task4]

        submission = Submission(
            work_base="parent_dir/",
            machine=machine,
            resources=resources,
            forward_common_files=["graph.pb"],
            backward_common_files=[],
            task_list=task_list,
        )
        submission.generate_jobs()
        str = machine.gen_script_header(submission.belonging_jobs[0])
        benchmark_str = textwrap.dedent(
            """\
            #!/bin/bash -l
            #SBATCH --parsable
            #SBATCH --nodes 1
            #SBATCH --ntasks-per-node 4
            #SBATCH --gres=gpu:2080Ti:2
            #SBATCH --partition GPU_2080Ti"""
        )
        self.assertEqual(str, benchmark_str)



import os
import sys
import unittest

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
__package__ = "tests"
from .context import RetrySignal, retry


class TestRetry(unittest.TestCase):
    def test_retry_fail(self):
        """Always retry"""

        @retry(max_retry=3, sleep=0.05, catch_exception=RetrySignal)
        def some_method():
            raise RetrySignal("Failed to do something")

        with self.assertRaises(RuntimeError):
            some_method()

    def test_retry_success(self):
        """Retry less than 3 times"""
        retry_times = [0]

        @retry(max_retry=3, sleep=0.05, catch_exception=RetrySignal)
        def some_method(retry_times):
            if retry_times[0] < 2:
                retry_times[0] += 1
                raise RetrySignal("Failed to do something")

        some_method(retry_times)



import os
import signal
import subprocess as sp

from dpdispatcher.base_context import BaseContext


class SPRetObj(object):
    def __init__(self, ret):
        self.data = ret

    def read(self):
        return self.data

    def readlines(self):
        lines = self.data.decode("utf-8").splitlines()
        ret = []
        for aa in lines:
            ret.append(aa + "\n")
        return ret


class LazyLocalContext(BaseContext):
    def __init__(
        self,
        local_root,
        remote_root=None,
        remote_profile={},
        *args,
        **kwargs,
    ):
        """
        local_root:
        remote_root:
        remote_profile:
        """
        assert type(local_root) == str
        self.init_local_root = local_root
        self.init_remote_root = remote_root
        self.temp_local_root = os.path.abspath(local_root)
        self.temp_remote_root = os.path.abspath(local_root)
        self.remote_profile = remote_profile
        # self.job_uuid = None
        # self.submission = None
        # if job_uuid:
        #    self.job_uuid=job_uuid
        # else:
        #    self.job_uuid = str(uuid.uuid4())

    @classmethod
    def load_from_dict(cls, context_dict):
        local_root = context_dict["local_root"]
        remote_root = context_dict.get("remote_root", None)
        remote_profile = context_dict.get("remote_profile", {})
        instance = cls(
            local_root=local_root,
            remote_root=remote_root,
            remote_profile=remote_profile,
        )
        return instance

    def bind_submission(self, submission):
        self.submission = submission
        self.local_root = os.path.join(self.temp_local_root, submission.work_base)
        self.remote_root = os.path.join(self.temp_local_root, submission.work_base)
        # dlog.debug("debug:LazyLocalContext.bind_submission;"
        #     "submission.submission_hash:{submission.submission_hash};"
        #     "self.local_root:{self.local_root};"
        #     "self.remote_root:{self.remote_root}")

    def get_job_root(self):
        return self.local_root

    def upload(
        self,
        jobs,
        # local_up_files,
        dereference=True,
    ):
        pass

    def download(
        self,
        jobs,
        # remote_down_files,
        check_exists=False,
        mark_failure=True,
        back_error=False,
    ):
        pass

    #    for ii in job_dirs :
    #        for jj in remote_down_files :
    #            fname = os.path.join(self.local_root, ii, jj)
    #            exists = os.path.exists(fname)
    #            if not exists:
    #                if check_exists:
    #                    if mark_failure:
    #                        with open(os.path.join(self.local_root, ii, 'tag_failure_download_%s' % jj), 'w') as fp: pass
    #                    else:
    #                        pass
    #                else:
    #                    raise RuntimeError('do not find download file ' + fname)

    def block_checkcall(self, cmd):
        # script_dir = os.path.join(self.local_root, self.submission.work_base)
        # os.chdir(script_dir)
        proc = sp.Popen(
            cmd, cwd=self.local_root, shell=True, stdout=sp.PIPE, stderr=sp.PIPE
        )
        o, e = proc.communicate()
        stdout = SPRetObj(o)
        stderr = SPRetObj(e)
        code = proc.returncode
        if code != 0:
            raise RuntimeError(
                "Get error code %d in locally calling %s with job: %s ",
                (code, cmd, self.submission.submission_hash),
            )
        return None, stdout, stderr

    def block_call(self, cmd):
        proc = sp.Popen(
            cmd, cwd=self.local_root, shell=True, stdout=sp.PIPE, stderr=sp.PIPE
        )
        o, e = proc.communicate()
        stdout = SPRetObj(o)
        stderr = SPRetObj(e)
        code = proc.returncode
        return code, None, stdout, stderr

    def clean(self):
        pass

    def write_file(self, fname, write_str):
        os.makedirs(self.remote_root, exist_ok=True)
        with open(os.path.join(self.remote_root, fname), "w") as fp:
            fp.write(write_str)

    def read_file(self, fname):
        with open(os.path.join(self.remote_root, fname), "r") as fp:
            ret = fp.read()
        return ret

    def check_file_exists(self, fname):
        # submission_work_base = os.path.join(self.local_root, self.submission.work_base)
        # file_to_be_checked = os.path.join(submission_work_base, fname)
        # print('debug:dpdispatcher.LazyLocalContext().check_file_exists:file_to_be_checked', file_to_be_checked)
        # return os.path.isfile(file_to_be_checked)
        return os.path.isfile(os.path.join(self.remote_root, fname))

    def call(self, cmd):
        cwd = os.getcwd()
        proc = sp.Popen(
            cmd, cwd=self.local_root, shell=True, stdout=sp.PIPE, stderr=sp.PIPE
        )
        return proc

    def kill(self, job_id):
        os.kill(job_id, signal.SIGTERM)

    def check_finish(self, proc):
        return proc.poll() != None

    def get_return(self, proc):
        ret = proc.poll()
        if ret is None:
            return None, None, None
        else:
            try:
                o, e = proc.communicate()
                stdout = SPRetObj(o)
                stderr = SPRetObj(e)
            except sp.SubprocessError:
                stdout = None
                stderr = None
        return ret, stdout, stderr



import glob
import json
import os
import shutil
import sys
import time
import unittest
import uuid

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
__package__ = "pbs"
from .context import PBS, JobStatus, LocalContext, LocalSession, setUpModule


@unittest.skipIf(not shutil.which("qsub"), "requires PBS")
class TestPBS(unittest.TestCase):
    def setUp(self):
        os.makedirs("loc", exist_ok=True)
        os.makedirs("rmt", exist_ok=True)
        os.makedirs("loc/task0", exist_ok=True)
        os.makedirs("loc/task1", exist_ok=True)
        for ii in ["loc/task0", "loc/task1"]:
            with open(os.path.join(ii, "test0"), "w") as fp:
                fp.write(str(uuid.uuid4()))
        work_profile = LocalSession({"work_path": "rmt"})
        self.ctx = LocalContext("loc", work_profile)
        self.pbs = PBS(self.ctx)

    def tearDown(self):
        shutil.rmtree("loc")
        shutil.rmtree("rmt")
        if os.path.exists("dpgen.log"):
            os.remove("dpgen.log")

    def test_gen_sub_script(self):
        job_dirs = ["task0", "task1"]
        self.pbs.context.upload(job_dirs, ["test0"])
        ret = self.pbs.sub_script(job_dirs, ["touch test1", "touch test2"])
        self.pbs.context.write_file("run.sub", ret)
        with open("run.sub", "w") as fp:
            fp.write(ret)

    # def test_sub_success(self) :
    #     job_dirs = ['task0', 'task1']
    #     self.pbs.context.upload(job_dirs, ['test0'])
    #     self.pbs.submit(job_dirs, ['touch test1', 'touch test2'])
    #     while True:
    #         ret = self.pbs.check_status()
    #         if ret == JobStatus.finished  :
    #             break
    #         time.sleep(1)
    #     self.assertTrue(os.path.isfile(os.path.join('rmt', self.pbs.context.remote_root, 'task0/tag_0_finished')))
    #     self.assertTrue(os.path.isfile(os.path.join('rmt', self.pbs.context.remote_root, 'task0/tag_1_finished')))
    #     self.assertTrue(os.path.isfile(os.path.join('rmt', self.pbs.context.remote_root, 'task1/tag_0_finished')))
    #     self.assertTrue(os.path.isfile(os.path.join('rmt', self.pbs.context.remote_root, 'task1/tag_1_finished')))
    #     self.assertTrue(os.path.isfile(os.path.join('rmt', self.pbs.context.remote_root, 'tag_finished')))
    #     self.assertTrue (os.path.isfile(os.path.join('rmt', self.pbs.context.remote_root, 'task0/test1')))
    #     self.assertTrue (os.path.isfile(os.path.join('rmt', self.pbs.context.remote_root, 'task1/test1')))
    #     self.assertTrue (os.path.isfile(os.path.join('rmt', self.pbs.context.remote_root, 'task0/test2')))
    #     self.assertTrue (os.path.isfile(os.path.join('rmt', self.pbs.context.remote_root, 'task1/test2')))

    # def test_sub_scancel(self) :
    #     job_dirs = ['task0', 'task1']
    #     self.pbs.context.upload(job_dirs, ['test0'])
    #     # sub
    #     self.pbs.submit(job_dirs, ['touch test1', 'sleep 10'])
    #     while True:
    #         ret = self.pbs.check_status()
    #         if ret == JobStatus.finished  :
    #             raise RuntimeError('should not finished')
    #         if ret == JobStatus.running :
    #             # wait for file writing
    #             time.sleep(2)
    #             job_id = self.pbs._get_job_id()
    #             os.system('scancel ' + job_id)
    #             break
    #         time.sleep(1)
    #     self.assertTrue (os.path.isfile(os.path.join('rmt', self.pbs.context.remote_root, 'task0/tag_0_finished')))
    #     self.assertFalse(os.path.isfile(os.path.join('rmt', self.pbs.context.remote_root, 'task0/tag_1_finished')))
    #     self.assertTrue (os.path.isfile(os.path.join('rmt', self.pbs.context.remote_root, 'task1/tag_0_finished')))
    #     self.assertFalse(os.path.isfile(os.path.join('rmt', self.pbs.context.remote_root, 'task1/tag_1_finished')))
    #     self.assertFalse(os.path.isfile(os.path.join('rmt', self.pbs.context.remote_root, 'tag_finished')))
    #     self.assertTrue (os.path.isfile(os.path.join('rmt', self.pbs.context.remote_root, 'task0/test1')))
    #     self.assertTrue (os.path.isfile(os.path.join('rmt', self.pbs.context.remote_root, 'task1/test1')))
    #     self.assertFalse(os.path.isfile(os.path.join('rmt', self.pbs.context.remote_root, 'task0/test2')))
    #     self.assertFalse(os.path.isfile(os.path.join('rmt', self.pbs.context.remote_root, 'task1/test2')))
    #     # sub restart
    #     self.pbs.submit(job_dirs, ['rm test1', 'touch test2'], restart = True)
    #     while True:
    #         ret = self.pbs.check_status()
    #         if ret == JobStatus.finished  :
    #             break
    #         time.sleep(1)
    #     self.assertTrue (os.path.isfile(os.path.join('rmt', self.pbs.context.remote_root, 'task0/tag_0_finished')))
    #     self.assertTrue (os.path.isfile(os.path.join('rmt', self.pbs.context.remote_root, 'task0/tag_1_finished')))
    #     self.assertTrue (os.path.isfile(os.path.join('rmt', self.pbs.context.remote_root, 'task1/tag_0_finished')))
    #     self.assertTrue (os.path.isfile(os.path.join('rmt', self.pbs.context.remote_root, 'task1/tag_1_finished')))
    #     self.assertTrue (os.path.isfile(os.path.join('rmt', self.pbs.context.remote_root, 'tag_finished')))
    #     self.assertTrue (os.path.isfile(os.path.join('rmt', self.pbs.context.remote_root, 'task0/test1')))
    #     self.assertTrue (os.path.isfile(os.path.join('rmt', self.pbs.context.remote_root, 'task1/test1')))
    #     self.assertTrue (os.path.isfile(os.path.join('rmt', self.pbs.context.remote_root, 'task0/test2')))
    #     self.assertTrue (os.path.isfile(os.path.join('rmt', self.pbs.context.remote_root, 'task1/test2')))
